/**
 * @class Fps
 */
var Fps = (function () {
    /**
     * @constructor
     */
    function Fps() {
        var _this = this;
        /**
         * Starts the measurement and calculation of the fps.
         */
        this.start = function (startTime) {
            _this.fpsFrameCounterPerSecond = 0;
            _this.fpsFrameTimeLastMeasurement = startTime;
            _this.fpsCurrentFpsValue = 0;
            _this.fpsLastUpdateTimeAbsolute = startTime;
        };
        /**
         * Update function that shall be called each frame.
         * Measures and calculates the timings and fps.
         * @return false if a maximum FPS limit is set and it is not enough time elapsed to handle the next frame,
         *         otherwise true
         */
        this.update = function (time) {
            /* limit maximum FPS if enabled */
            if (_this.fpsMaxAllowed != 0) {
                /* If not enough time has passed for current frame, return immediately without update values below! */
                if (time < _this.fpsLastUpdateTimeAbsolute + (1000 / _this.fpsMaxAllowed)) {
                    return false;
                }
            }
            /* calculate fps */
            if (time > _this.fpsFrameTimeLastMeasurement + 1000) {
                _this.fpsCurrentFpsValue = _this.fpsFrameCounterPerSecond;
                _this.fpsFrameCounterPerSecond = 0;
                _this.fpsFrameTimeLastMeasurement = time;
            }
            /* update frame counter and store absolute time of this call */
            _this.fpsFrameCounterPerSecond++;
            _this.fpsLastUpdateTimeAbsolute = time;
            return true;
        };
        this.getCurrentFps = function () {
            return _this.fpsCurrentFpsValue;
        };
        this.getMaxAllowedFps = function () {
            return _this.fpsMaxAllowed;
        };
        this.setMaxAllowedFps = function (maxAllowedFps) {
            _this.fpsMaxAllowed = maxAllowedFps;
        };
        this.fpsMaxAllowed = 120;
    }
    return Fps;
})();
/**
 * @class GraphicsLoop
 */
var GraphicsLoop = (function () {
    function GraphicsLoop(htmlCvsElem, useDoubleBuffer) {
        var _this = this;
        if (useDoubleBuffer === void 0) { useDoubleBuffer = true; }
        this.backgroundColor = undefined;
        this.update = function (time) {
        };
        this.draw = function (ctx, time) {
        };
        this.startMainLoop = function () {
            if (!_this.loopRunning) {
                window.requestAnimationFrame(function (time) {
                    _this.fps.start(time);
                    _this.lastFrameTimeAbsolute = time;
                    _this.loopRunning = true;
                    _this.animationFrameId = window.requestAnimationFrame(_this.mainLoop);
                });
            }
        };
        this.stopMainLoop = function () {
            if (_this.loopRunning) {
                _this.stopLoopRequested = true;
            }
        };
        this.mainLoop = function (time) {
            var deltaFrameTime;
            deltaFrameTime = time - _this.lastFrameTimeAbsolute;
            if (_this.fps.update(time) === false) {
                requestAnimationFrame(_this.mainLoop);
                return;
            }
            /* call update function */
            _this.update(deltaFrameTime);
            /* prepare canvas and background, call draw function */
            if (_this.backgroundColor) {
                var fillStyleBackup = _this.ctxUser.fillStyle;
                _this.ctxUser.fillStyle = _this.backgroundColor;
                _this.ctxUser.fillRect(0, 0, _this.canvasMain.width, _this.canvasMain.height);
                _this.ctxUser.fillStyle = fillStyleBackup;
            }
            else {
                _this.ctxUser.clearRect(0, 0, _this.canvasMain.width, _this.canvasMain.height);
            }
            _this.draw(_this.ctxUser, deltaFrameTime);
            if (_this.useDoubleBuffering) {
                _this.ctxMain.drawImage(_this.canvasDoubleBuffer, 0, 0);
            }
            /* draw FPS value if enabled */
            if (_this.FpsDrawFpsValue) {
                _this.drawFPS();
            }
            /* save absolute time of this frame */
            _this.lastFrameTimeAbsolute = time;
            /* prepare next loop and pause */
            if (_this.stopLoopRequested) {
                window.cancelAnimationFrame(_this.animationFrameId);
                _this.stopLoopRequested = false;
                _this.loopRunning = false;
            }
            else {
                _this.animationFrameId = window.requestAnimationFrame(_this.mainLoop);
            }
        };
        this.drawFPS = function () {
            _this.ctxMain.save();
            _this.ctxMain.font = '10px Arial';
            _this.ctxMain.fillStyle = _this.FpsFillStyle;
            _this.ctxMain.fillText("FPS: " + _this.fps.getCurrentFps(), _this.canvasMain.width - 60, 20);
            _this.ctxMain.restore();
        };
        this.getCurrentFps = function () {
            return _this.fps.getCurrentFps();
        };
        this.setMaxAllowedFps = function (fps) {
            _this.fps.setMaxAllowedFps(fps);
        };
        this.setBackground = function (color) {
            _this.backgroundColor = color;
        };
        this.getCanvasWidth = function () {
            return _this.canvasMain.width;
        };
        this.getCanvasHeight = function () {
            return _this.canvasMain.height;
        };
        this.setFullScreen = function () {
            if (_this.canvasMain.webkitRequestFullScreen) {
                _this.canvasMain.webkitRequestFullScreen();
            }
            else {
                if (_this.canvasMain['mozRequestFullScreen']) {
                    _this.canvasMain['mozRequestFullScreen']();
                }
                else if (_this.canvasMain['msRequestFullscreen']) {
                    _this.canvasMain['msRequestFullscreen']();
                }
            }
        };
        this.canvasMain = htmlCvsElem;
        this.ctxMain = this.canvasMain.getContext("2d");
        this.useDoubleBuffering = useDoubleBuffer;
        if (this.useDoubleBuffering) {
            this.canvasDoubleBuffer = document.createElement('canvas');
            this.canvasDoubleBuffer.width = this.canvasMain.width;
            this.canvasDoubleBuffer.height = this.canvasMain.height;
            this.ctxDoubleBuffer = this.canvasDoubleBuffer.getContext("2d");
            this.ctxUser = this.ctxDoubleBuffer;
        }
        else {
            this.ctxUser = this.ctxMain;
        }
        this.fps = new Fps();
        this.FpsDrawFpsValue = false;
        this.FpsFillStyle = '#000000';
        this.loopRunning = false;
        this.stopLoopRequested = false;
    }
    return GraphicsLoop;
})();
//# sourceMappingURL=GraphicsLoop.js.map